//Converges on TESLA K20. Simulation was run overnight! 
//May NOT run on a budget card (too many threads and blocks)!
//512*512*1000 paths, 10000 trades per path, gradient step U[-0.005, 0.005] per single coordinate.
//The starting value is: equal amout of capital among all assets.

#include "cuda_runtime.h"
#include "device_launch_parameters.h"

#include <stdio.h>
#include <stdlib.h>
#include <cuda.h>
#include <curand.h>
#include <curand_kernel.h>
#include <thrust/random/linear_congruential_engine.h>
#include <thrust/random/normal_distribution.h>

#include <thrust/host_vector.h>
#include <thrust/device_vector.h>
#include <thrust/sort.h>
#include <thrust/copy.h>
#include <thrust/random.h>
#include <thrust/iterator/counting_iterator.h>
#include <thrust/functional.h>
#include <thrust/transform_reduce.h>

#include <random>
#include <iostream>
#include <algorithm>
#include <vector>
#include <cmath>
#include <set>
#include <ostream>
#include <fstream>

#include <time.h>

#include<omp.h>
#include<ppl.h>


//standard hash function
__host__ __device__ unsigned int hash(unsigned int a)
{
    a = (a+0x7ed55d16) + (a<<12);
    a = (a^0xc761c23c) ^ (a>>19);
    a = (a+0x165667b1) + (a<<5);
    a = (a+0xd3a2646c) ^ (a<<9);
    a = (a+0xfd7046c5) + (a<<3);
    a = (a^0xb55a4f09) ^ (a>>16);
    return a;
}

//In practice the daily returns are (nearly) always in [-50%, 50%]
__host__ __device__  double cutTails(double x)
{
    if( x > 0.5 ) x = 0.5;
	if( x < -0.5) x = -0.5;
    return x;
}

__global__ void logUtilForFrac(double f1, double f2, double f3, double f4, double f5, double f6, double f7, double* res)
{
	long 	tid = threadIdx.x + blockIdx.x * blockDim.x;
	double f8 = 1.0 - f1 - f2 -f3 - f4 - f5 -f6 - f7; //the residual capital fraction, invested in riskfree bonds

	//expected daily returns
	double meanRets[7] = {0.0004556028, 0.0002526534, 0.0002631352, 0.0002793451, 0.0003849809, 0.0002089822, 0.0002541450};
	//embedded Cholesky decomposition of the covariance matrix
	double c1[7] =  {0.036936230,  0.0000000000,  0.000000000, 0.0000000000,  0.000000e+00, 0.000000000, 0.00000000};
	double c2[7] =  {0.003072113,  0.0183089921,  0.000000000, 0.0000000000,  0.000000e+00, 0.000000000, 0.00000000};
	double c3[7] =  {0.002933369,  0.0074403765,  0.016156070, 0.0000000000,  0.000000e+00, 0.000000000, 0.00000000};
	double c4[7] =  {0.006934453,  0.0065668130,  0.006447719, 0.0139242061,  0.000000e+00, 0.000000000, 0.00000000};
	double c5[7] =  {0.025927277, -0.0009107914, -0.002605173, 0.0002330022,  1.450277e-02, 0.000000000, 0.00000000};
	double c6[7] =  {0.002162088,  0.0072983969,  0.004280576, 0.0023632050, -1.593418e-05, 0.012981916, 0.00000000};
	double c7[7] =  {0.004010336,  0.0084436921,  0.007796282, 0.0034809939, -1.408359e-03, 0.002561217, 0.01275303};

	double wealth = 1.0;
	double riskFreeReturn= 1.0 + (0.04 / 365);

	for( int i=0; i<10000; ++i )
	{
		//do simulation
		unsigned long seed = hash(tid) + clock();
		thrust::minstd_rand rng(seed);
		thrust::random::experimental::normal_distribution<double> dist(0.0, 1.0);
		double r1 = dist(rng);
		double r2 = dist(rng);
		double r3 = dist(rng);
		double r4 = dist(rng);
		double r5 = dist(rng);
		double r6 = dist(rng);
		double r7 = dist(rng);

		double d1 = 1.0 + cutTails( meanRets[0] + c1[0]*r1 );
		double d2 = 1.0 + cutTails( meanRets[1] + c2[0]*r1 + c2[1]*r2 );
		double d3 = 1.0 + cutTails( meanRets[2] + c3[0]*r1 + c3[1]*r2 + c3[2]*r3 );
		double d4 = 1.0 + cutTails( meanRets[3] + c4[0]*r1 + c4[1]*r2 + c4[2]*r3 + c4[3]*r4 );
		double d5 = 1.0 + cutTails( meanRets[4] + c5[0]*r1 + c5[1]*r2 + c5[2]*r3 + c5[3]*r4 + c5[4]*r5 );
		double d6 = 1.0 + cutTails( meanRets[5] + c6[0]*r1 + c6[1]*r2 + c6[2]*r3 + c6[3]*r4 + c6[4]*r5 + c6[5]*r6 );
		double d7 = 1.0 + cutTails( meanRets[6] + c7[0]*r1 + c7[1]*r2 + c7[2]*r3 + c7[3]*r4 + c7[4]*r5 + c7[5]*r6 + c7[6]*r7 );


		wealth *= ( f1*d1 + f2*d2 + f3*d3 + f4*d4 + f5*d5 + f6*d6 + f7*d7 + f8*riskFreeReturn );
	};

	//end simulation
	res[tid] = log(wealth);
}

void sumulatePathHost(double f1, double f2, double f3, double f4, double f5, double f6, double f7, unsigned long s, double& res)
{
	double f8 = 1.0 - f1 - f2 -f3 - f4 - f5 -f6 - f7;
	//expected daily returns
	double meanRets[7] = {0.0004556028, 0.0002526534, 0.0002631352, 0.0002793451, 0.0003849809, 0.0002089822, 0.0002541450};
	//embedded Cholesky decomposition of the covariance matrix
	double c1[7] =  {0.036936230,  0.0000000000,  0.000000000, 0.0000000000,  0.000000e+00, 0.000000000, 0.00000000};
	double c2[7] =  {0.003072113,  0.0183089921,  0.000000000, 0.0000000000,  0.000000e+00, 0.000000000, 0.00000000};
	double c3[7] =  {0.002933369,  0.0074403765,  0.016156070, 0.0000000000,  0.000000e+00, 0.000000000, 0.00000000};
	double c4[7] =  {0.006934453,  0.0065668130,  0.006447719, 0.0139242061,  0.000000e+00, 0.000000000, 0.00000000};
	double c5[7] =  {0.025927277, -0.0009107914, -0.002605173, 0.0002330022,  1.450277e-02, 0.000000000, 0.00000000};
	double c6[7] =  {0.002162088,  0.0072983969,  0.004280576, 0.0023632050, -1.593418e-05, 0.012981916, 0.00000000};
	double c7[7] =  {0.004010336,  0.0084436921,  0.007796282, 0.0034809939, -1.408359e-03, 0.002561217, 0.01275303};
		
	double riskFreeReturn= 1.0 + (0.04 / 365);

	double wealth = 1.0;
	for( int i=0; i<100; ++i )
	{
		//do simulation
		unsigned long seed = hash(s) + clock();
   		std::minstd_rand rng;
		rng.seed(seed);
        std::normal_distribution<double>dist(0.0, 1.0);

		double r1 = dist(rng);
		double r2 = dist(rng);
		double r3 = dist(rng);
		double r4 = dist(rng);
		double r5 = dist(rng);
		double r6 = dist(rng);
		double r7 = dist(rng);

		double d1 = 1.0 + cutTails( meanRets[0] + c1[0]*r1 );
		double d2 = 1.0 + cutTails( meanRets[1] + c2[0]*r1 + c2[1]*r2 );
		double d3 = 1.0 + cutTails( meanRets[2] + c3[0]*r1 + c3[1]*r2 + c3[2]*r3 );
		double d4 = 1.0 + cutTails( meanRets[3] + c4[0]*r1 + c4[1]*r2 + c4[2]*r3 + c4[3]*r4 );
		double d5 = 1.0 + cutTails( meanRets[4] + c5[0]*r1 + c5[1]*r2 + c5[2]*r3 + c5[3]*r4 + c5[4]*r5 );
		double d6 = 1.0 + cutTails( meanRets[5] + c6[0]*r1 + c6[1]*r2 + c6[2]*r3 + c6[3]*r4 + c6[4]*r5 + c6[5]*r6 );
		double d7 = 1.0 + cutTails( meanRets[6] + c7[0]*r1 + c7[1]*r2 + c7[2]*r3 + c7[3]*r4 + c7[4]*r5 + c7[5]*r6 + c7[6]*r7 );


		wealth *= ( f1*d1 + f2*d2 + f3*d3 + f4*d4 + f5*d5 + f6*d6 + f7*d7 + f8*riskFreeReturn );
	};

	res = log(wealth);
}

//to compare the results from CPU and GPU
 double logUtilForFracHost(double f1, double f2, double f3, double f4, double f5, double f6, double f7, long nPathes)
{	
	std::vector<double> logTermWealth( nPathes );
	Concurrency::parallel_for(size_t(0), (size_t)nPathes, [&] (size_t s)
	{
		sumulatePathHost(f1, f2, f3, f4, f5, f6, f7, s, logTermWealth[s]);
	});
	
	double meanLogWealth = 0.0;
	for( long i=0; i<nPathes; ++i ) 
	{
			meanLogWealth += logTermWealth[i];
	};
	meanLogWealth /= nPathes;

	return(meanLogWealth);	
}



int main()
{
	cudaDeviceProp devProp;
    cudaGetDeviceProperties(&devProp, 0);
	cudaError_t cudaStatus = cudaSetDevice(0);
	const long BLOCKs  = 512; 
	const long THREADs = 512;
	const long resLen = BLOCKs* THREADs;
	double * res = (double*)malloc(resLen*sizeof(double)); 
	double *dev_res = 0;
	cudaStatus = cudaMalloc((void**)&dev_res, resLen * sizeof(double));
	
	const long N_SIMULATIONS = 300;
	double f1, f2, f3, f4, f5, f6, f7, f8;
	double h1, h2, h3, h4, h5, h6, h7, h8; //fractions to achieve maximum meanLogWealth
	f1 = f2 = f3 = f4 = f5 = f6 = f7 = f8 = h1 = h2 = h3 = h4 = h5 = h6 = h7 = h8 = 1.0 / 8.0;
	/*restart the 2nd iteration from the best result of the 2nd iteration*/
	//double rstVls[8] = {0.0261616, 0.159418, 0.208649, 0.118518, 0.215925, 0.032894,  0.0981397, 0.140295};
	/*  double rstVls[8] = {0.0160567, 0.156136, 0.226763, 0.14866,  0.229188, 0.0116993, 0.0929747, 0.118522};
	f1 = h1 = rstVls[0];
	f2 = h2 = rstVls[1];
	f3 = h3 = rstVls[2];
	f4 = h4 = rstVls[3];
	f5 = h5 = rstVls[4];
	f6 = h6 = rstVls[5];
	f7 = h7 = rstVls[6];
	f8 = h8 = rstVls[7];*/
	double maxMeanLogWealth = 0.0;


	for(long simulation = 0; simulation<N_SIMULATIONS; ++simulation)
	{
		//random increments
		std::default_random_engine generator((unsigned int)time(0));
		std::uniform_real_distribution<double> distribution(-0.005,0.005); //step decreased from (-0.01, 0.01) to (-0.005,0.005)
		f1 = std::max<double>(0.0, f1 + distribution(generator));
		f2 = std::max<double>(0.0, f2 + distribution(generator));
		f3 = std::max<double>(0.0, f3 + distribution(generator));
		f4 = std::max<double>(0.0, f4 + distribution(generator));
		f5 = std::max<double>(0.0, f5 + distribution(generator));
		f6 = std::max<double>(0.0, f6 + distribution(generator));
		f7 = std::max<double>(0.0, f7 + distribution(generator));
		f8 = std::max<double>(0.0, f8 + distribution(generator));
		
		double radius = f1 + f2 + f3 + f4 + f5 + f6 + f7 + f8;
		f1 /= radius;
		f2 /= radius;
		f3 /= radius;
		f4 /= radius;
		f5 /= radius;
		f6 /= radius;
		f7 /= radius;
		f8 /= radius; 


		/* Many budget NVIDIA cards can withstand maximum (512 Blocks * 512 Threads). Tricking with dim3 does not help,
		   in total it still must be 512^2 parallel entries. But we can repeatedly call the simulation onDevice.*/   
		const long ext = 100; //how many times to repeat the simulation onDevice
		double meanLogWealth = 0.0;
		for( int iter=0; iter<ext; ++iter )
		{
			logUtilForFrac<<<BLOCKs,THREADs>>>((double)f1, (double)f2, (double)f3, (double)f4, (double)f5, (double)f6, (double)f7, dev_res);
			cudaStatus = cudaDeviceSynchronize();
			if (cudaStatus != cudaSuccess) fprintf(stderr, "cudaDeviceSynchronize failed!");
			cudaStatus = cudaMemcpy(res, dev_res, resLen * sizeof(double), cudaMemcpyDeviceToHost);
			if (cudaStatus != cudaSuccess) fprintf(stderr, "cudaMemcpye failed!");

			double meanLogWealthTmp = 0.0;
			for( long i=0; i<resLen; ++i ) 
					meanLogWealthTmp += res[i];
		
			meanLogWealth += meanLogWealthTmp;			
		}
		
		meanLogWealth /= ((double)resLen * (double)ext); //if not cast to double, may occure the overflow of long

		std::cout << simulation << " : " << meanLogWealth << " : " <<  f1 << " : " <<  f2 << " : " <<  f3 
					<< " : " <<  f4 << " : " <<  f5 << " : " <<  f6 << " : " <<  f7 << " : " << f8 << "\n";

		//NB! Turn of when starting overnight simulation, otherwise it will be too slow
		//double meanLogWealthHost = logUtilForFracHost(f1, f2, f3, f4, f5, f6, f7, resLen*ext);
	    		
		if(meanLogWealth > maxMeanLogWealth)
		{
			maxMeanLogWealth = meanLogWealth;
			h1 = f1;
			h2 = f2;
			h3 = f3;
			h4 = f4;
			h5 = f5;
			h6 = f6;
			h7 = f7;
			h8 = f8;
		};

		f1 = h1;
		f2 = h2;
		f3 = h3;
		f4 = h4;
		f5 = h5;
		f6 = h6;
		f7 = h7;
		f8 = h8;
			
	}

	cudaStatus = cudaDeviceReset();
    if (cudaStatus != cudaSuccess) fprintf(stderr, "cudaDeviceReset failed!");
    
	free(res);
	cudaFree(dev_res);
    return 0;
}


